"""
mwrap_typecheck.py — Semantic analysis and type classification.

Copyright (c) 2007-2008  David Bindel
See the file COPYING for copying permissions

Converted to Python by Zydrunas Gimbutas (2026),
with assistance from Claude Code / Claude Opus 4.6 (Anthropic).
"""

import sys
from mwrap_ast import (
    VT, Expr, TypeQual, Var, Func,
    promote_int,
    iospec_is_input, iospec_is_output, iospec_is_inonly,
)


# ---------------------------------------------------------------------------
# Label assignment
# ---------------------------------------------------------------------------

def _label_dim_args_expr(args, icount):
    """Assign input_label to dimension expressions; returns updated icount."""
    for e in args:
        e.input_label = icount
        icount += 1
    return icount


def _label_dim_args_var(vars, icount):
    """Walk var list, label dimension expressions."""
    for v in vars:
        if v.qual:
            icount = _label_dim_args_expr(v.qual.args, icount)
    return icount


def _label_args_var(vars, icount, ocount):
    """Assign input/output labels to vars; returns (icount, ocount)."""
    for v in vars:
        if iospec_is_input(v.iospec):
            v.input_label = icount
            icount += 1
        if iospec_is_output(v.iospec):
            v.output_label = ocount
            ocount += 1
    return icount, ocount


def label_args(f):
    """Assign prhs[] / plhs[] indices to a Func's variables."""
    icount = 1 if f.thisv else 0
    ocount = 0
    icount, ocount = _label_args_var(f.ret, icount, ocount)
    icount, ocount = _label_args_var(f.args, icount, ocount)
    icount = _label_dim_args_var(f.ret, icount)
    icount = _label_dim_args_var(f.args, icount)


# ---------------------------------------------------------------------------
# Type-info assignment
# ---------------------------------------------------------------------------

def _assign_scalar_tinfo(v, line, tags, tagp, tagr, taga, tagar):
    """Assign tinfo for a scalar/complex type. Returns error count."""
    if not v.qual:
        v.tinfo = tags
    elif v.qual.qual == '*':
        v.tinfo = tagp
    elif v.qual.qual == '&':
        v.tinfo = tagr
    elif v.qual.qual == 'a':
        v.tinfo = taga
        # check max 2D
        if len(v.qual.args) > 2:
            print(f"Error ({line}): Array {v.name} should be 1D or 2D",
                  file=sys.stderr)
            return 1
    elif v.qual.qual == 'r':
        v.tinfo = tagar
        if tagar == VT.unk:
            print(f"Error ({line}): Array ref {v.name} must be to a real array",
                  file=sys.stderr)
            return 1
        if len(v.qual.args) > 2:
            print(f"Error ({line}): Array {v.name} should be 1D or 2D",
                  file=sys.stderr)
            return 1
    else:
        assert False, f"Unknown qual '{v.qual.qual}'"
    return 0


def assign_tinfo(ctx, v, line):
    """Assign VT_* tinfo to a single Var. Returns error count."""
    bt = v.basetype

    if ctx.is_scalar_type(bt):
        return _assign_scalar_tinfo(v, line,
                                    VT.scalar, VT.p_scalar, VT.r_scalar,
                                    VT.array, VT.rarray)
    elif ctx.is_cscalar_type(bt):
        return _assign_scalar_tinfo(v, line,
                                    VT.cscalar, VT.p_cscalar, VT.r_cscalar,
                                    VT.carray, VT.unk)
    elif ctx.is_zscalar_type(bt):
        return _assign_scalar_tinfo(v, line,
                                    VT.zscalar, VT.p_zscalar, VT.r_zscalar,
                                    VT.zarray, VT.unk)
    elif bt == "const":
        if v.qual:
            print(f"Error ({line}): Constant {v.name} cannot have modifiers",
                  file=sys.stderr)
            return 1
        v.tinfo = VT.const
        # Strip quotes from name if present
        if v.name and v.name.startswith("'"):
            v.name = v.name.replace("'", "")

    elif bt == "cstring":
        if v.qual and v.qual.qual != 'a':
            print(f"Error ({line}): String type {v.name} cannot have modifiers",
                  file=sys.stderr)
            return 1
        if v.qual and len(v.qual.args) > 1:
            print(f"Error ({line}): Strings are one dimensional",
                  file=sys.stderr)
            return 1
        v.tinfo = VT.string

    elif bt == "mxArray":
        if v.qual:
            print(f"Error ({line}): mxArray {v.name} cannot have modifiers",
                  file=sys.stderr)
            return 1
        v.tinfo = VT.mx

    else:
        # Object type
        if not v.qual:
            v.tinfo = VT.obj
        elif v.qual.qual == '*':
            v.tinfo = VT.p_obj
        elif v.qual.qual == '&':
            v.tinfo = VT.r_obj
        elif v.qual.qual in ('a', 'r'):
            print(f"Error ({line}): {v.name} cannot be an array of object {bt}",
                  file=sys.stderr)
            return 1
        else:
            assert False

    return 0


# ---------------------------------------------------------------------------
# Return-value validation
# ---------------------------------------------------------------------------

def _typecheck_return(ctx, ret, line):
    if not ret:
        return 0
    v = ret[0]
    err = assign_tinfo(ctx, v, line)

    if v.tinfo in (VT.array, VT.carray, VT.zarray):
        if not (v.qual and v.qual.args):
            print(f"Error ({line}): Return array {v.name} must have dims",
                  file=sys.stderr)
            err += 1
    elif v.tinfo == VT.const:
        print(f"Error ({line}): Cannot return constant", file=sys.stderr)
        err += 1
    elif v.tinfo == VT.rarray:
        print(f"Error ({line}): Ref to array {v.name} looks just like array on return",
              file=sys.stderr)
        err += 1
    elif v.tinfo == VT.string and v.qual:
        print(f"Error ({line}): Return string {v.name} cannot have dims",
              file=sys.stderr)
        err += 1
    return err


# ---------------------------------------------------------------------------
# Argument validation
# ---------------------------------------------------------------------------

def _typecheck_args(ctx, args, line):
    err = 0
    for v in args:
        err += assign_tinfo(ctx, v, line)

        if iospec_is_inonly(v.iospec):
            continue

        # Output / inout checks
        if v.name and v.name[0:1].isdigit():
            print(f"Error ({line}): Number {v.name} cannot be output",
                  file=sys.stderr)
            err += 1

        if (v.tinfo in (VT.obj, VT.p_obj, VT.r_obj) and
                not ctx.is_mxarray_type(v.basetype)):
            print(f"Error ({line}): Object {v.name} cannot be output",
                  file=sys.stderr)
            err += 1
        elif (v.tinfo in (VT.array, VT.carray, VT.zarray, VT.rarray) and
              v.iospec == 'o' and
              not (v.qual and v.qual.args)):
            print(f"Error ({line}): Output array {v.name} must have dims",
                  file=sys.stderr)
            err += 1
        elif v.tinfo == VT.rarray and not iospec_is_output(v.iospec):
            print(f"Error ({line}): Array ref {v.name} *must* be output",
                  file=sys.stderr)
            err += 1
        elif v.tinfo == VT.scalar:
            print(f"Error ({line}): Scalar {v.name} cannot be output",
                  file=sys.stderr)
            err += 1
        elif v.tinfo == VT.const:
            print(f"Error ({line}): Constant {v.name} cannot be output",
                  file=sys.stderr)
            err += 1
        elif v.tinfo == VT.string and not (v.qual and v.qual.args):
            print(f"Error ({line}): String {v.name} cannot be output without size",
                  file=sys.stderr)
            err += 1
        elif v.tinfo == VT.mx and v.iospec == 'b':
            print(f"Error ({line}): mxArray {v.name} cannot be used for inout",
                  file=sys.stderr)
            err += 1

    return err


# ---------------------------------------------------------------------------
# Fortran-ize arguments
# ---------------------------------------------------------------------------

def _fortranize_args_var(args, line):
    """Convert scalar args to pointer-to-scalar for FORTRAN. Returns error count."""
    err = 0
    for v in args:
        if v.tinfo in (VT.obj, VT.p_obj, VT.r_obj):
            print(f"Error ({line}): Cannot pass object {v.name} to FORTRAN",
                  file=sys.stderr)
            err += 1
        elif v.tinfo == VT.rarray:
            print(f"Error ({line}): Cannot pass pointer ref {v.name} to FORTRAN",
                  file=sys.stderr)
            err += 1
        elif v.tinfo == VT.string:
            print(f"Warning ({line}): Danger passing C string {v.name} to FORTRAN",
                  file=sys.stderr)
        elif v.tinfo in (VT.scalar, VT.r_scalar):
            v.tinfo = VT.p_scalar
        elif v.tinfo in (VT.cscalar, VT.r_cscalar):
            v.tinfo = VT.p_cscalar
        elif v.tinfo in (VT.zscalar, VT.r_zscalar):
            v.tinfo = VT.p_zscalar
    return err


def _fortranize_ret(v, line):
    if not v:
        return 0
    if v.tinfo in (VT.cscalar, VT.zscalar):
        print(f"Warning ({line}): Danger returning complex from FORTRAN",
              file=sys.stderr)
    elif v.tinfo != VT.scalar:
        print(f"Error ({line}): Can only return scalars from FORTRAN",
              file=sys.stderr)
        return 1
    return 0


def _fortranize_args(f, line):
    if not f.fort:
        return 0
    err = _fortranize_args_var(f.args, line)
    if f.ret:
        err += _fortranize_ret(f.ret[0], line)
    return err


# ---------------------------------------------------------------------------
# Top-level typecheck
# ---------------------------------------------------------------------------

def typecheck(ctx, f, line):
    """Run full semantic analysis on a Func. Returns error count."""
    label_args(f)
    return (_typecheck_return(ctx, f.ret, line) +
            _typecheck_args(ctx, f.args, line) +
            _fortranize_args(f, line))



"""
mwrap_mgen.py — MATLAB stub generator.

Copyright (c) 2007-2008  David Bindel
See the file COPYING for copying permissions

Converted to Python by Zydrunas Gimbutas (2026),
with assistance from Claude Code / Claude Opus 4.6 (Anthropic).
"""

from mwrap_ast import VT, id_string


def _output_arg_names(args):
    """Collect names of output/inout arguments."""
    return [v.name for v in args if v.iospec in ('o', 'b')]


def _input_arg_strs(args):
    """Collect input argument strings (with leading ', ')."""
    parts = []
    for v in args:
        if v.tinfo == VT.const:
            parts.append(", 0")
        elif v.iospec in ('i', 'b'):
            parts.append(f", {v.name}")
    return parts


def _dim_arg_strs(vars):
    """Collect dimension argument strings (with leading ', ')."""
    parts = []
    for v in vars:
        if v.qual:
            for e in v.qual.args:
                parts.append(f", {e.value}")
    return parts


def print_matlab_call(fp, ctx, f, mexfunc):
    """Emit MATLAB stub for function call f."""
    fp.write(f"mex_id_ = '{id_string(ctx, f)}';\n")

    out_names = _output_arg_names(f.ret) + _output_arg_names(f.args)
    if out_names:
        fp.write(f"[{', '.join(out_names)}] = ")

    rhs = [f"mex_id_"]
    if f.thisv:
        rhs.append(f", {f.thisv}")
    rhs.extend(_input_arg_strs(f.args))
    rhs.extend(_dim_arg_strs(f.ret))
    rhs.extend(_dim_arg_strs(f.args))
    fp.write(f"{mexfunc}({''.join(rhs)});\n")



"""
mwrap_lexer.py — Line-oriented lexer for .mw files.

Copyright (c) 2007-2008  David Bindel
See the file COPYING for copying permissions

Converted to Python by Zydrunas Gimbutas (2026),
with assistance from Claude Code / Claude Opus 4.6 (Anthropic).
"""

import re
import sys
import os
from enum import Enum, auto
from dataclasses import dataclass
from typing import Optional, TextIO, List


class TokenType(Enum):
    ID        = auto()
    NUMBER    = auto()
    STRING    = auto()
    NEW       = auto()
    FORTRAN   = auto()
    INPUT     = auto()
    OUTPUT    = auto()
    INOUT     = auto()
    CLASS     = auto()
    TYPEDEF   = auto()
    CPU       = auto()
    GPU       = auto()
    PUNCT     = auto()      # single characters: ( ) , ; * & [ ] . - > = :
    NON_C_LINE = auto()
    EOF       = auto()


KEYWORDS = {
    "new":      TokenType.NEW,
    "FORTRAN":  TokenType.FORTRAN,
    "input":    TokenType.INPUT,
    "output":   TokenType.OUTPUT,
    "inout":    TokenType.INOUT,
    "class":    TokenType.CLASS,
    "typedef":  TokenType.TYPEDEF,
    "cpu":      TokenType.CPU,
    "gpu":      TokenType.GPU,
}

# Regex for tokenising a '#' line body
_TOKEN_RE = re.compile(
    r"(//[^\n]*)"               # comment (rest of line) — skip
    r"|('[^'\n]*'?)"            # string literal (single-quoted)
    r"|((?:::)?[_a-zA-Z][_a-zA-Z0-9]*(?:::(?:[_a-zA-Z][_a-zA-Z0-9]*))*)"  # ID (may have ::)
    r"|([0-9]+)"                # number
    r"|([->()\[\],;*&=:.])"     # punctuation
    r"|[ \t\r]+"                # whitespace — skip
)


@dataclass
class Token:
    type: TokenType
    value: str
    line: int


def _is_name_char(c):
    return c.isalnum() or c == '_'


def _fname_scan_line(text):
    """Extract function name from '@function ...' tail, return name.m."""
    paren = text.find('(')
    if paren < 0:
        paren = len(text)
    # Walk back from paren to find last alnum
    end = paren
    while end > 0 and not _is_name_char(text[end - 1]):
        end -= 1
    start = end
    while start > 0 and _is_name_char(text[start - 1]):
        start -= 1
    name = text[start:end]
    return name + ".m"


class Lexer:
    """Line-oriented lexer for .mw files.

    Usage:
        lex = Lexer(outfp, outcfp, mbatching_flag, listing_flag)
        for tok in lex.lex_file(filename):
            ...   # only tokens from '#' lines; NON_C_LINE for other lines
    """

    def __init__(self, outfp=None, outcfp=None,
                 mbatching_flag=False, listing_flag=False):
        self.outfp: Optional[TextIO] = outfp
        self.outcfp: Optional[TextIO] = outcfp
        self.mbatching_flag: bool = mbatching_flag
        self.listing_flag: bool = listing_flag
        self.linenum: int = 0
        self.current_ifname: str = ""

        # File include stack
        self._file_stack: List = []          # [(fp, linenum, ifname), ...]
        self._current_fp: Optional[TextIO] = None

    # ------------------------------------------------------------------
    # public interface
    # ------------------------------------------------------------------

    def lex_file(self, filename):
        """Yield Token objects from *filename*."""
        fp = open(filename, "r")
        self._current_fp = fp
        self.current_ifname = filename
        self.linenum = 1
        yield from self._lex_stream()

    # ------------------------------------------------------------------
    # directive handlers
    # ------------------------------------------------------------------

    def _handle_block_c(self, line):
        """Process a line in block C mode. Returns False when block ends."""
        stripped = line.rstrip('\r\n')
        if re.match(r'^\$\][ \t\r]*$', stripped):
            self.linenum += 1
            return False
        if self.outcfp:
            self.outcfp.write(line)
        self.linenum += 1
        return True

    def _handle_comment(self):
        """Handle // comment line."""
        self.linenum += 1
        yield Token(TokenType.NON_C_LINE, "", self.linenum - 1)

    def _handle_function(self, stripped):
        """Handle @function directive."""
        tail = stripped[len("@function"):]
        tail = tail.rstrip('\r\n')
        fname = _fname_scan_line(tail)
        if self.mbatching_flag:
            if self.outfp:
                self.outfp.close()
            try:
                self.outfp = open(fname, "w")
            except OSError:
                print(f"Error: Could not write {fname}",
                      file=sys.stderr)
                sys.exit(1)
        if self.listing_flag:
            print(fname)
        if self.outfp:
            self.outfp.write(f"function{tail}\n")
        self.linenum += 1
        yield Token(TokenType.NON_C_LINE, "", self.linenum - 1)

    def _handle_include(self, stripped):
        """Handle @include directive. Pushes current file onto stack."""
        rest = stripped[len("@include"):].strip().rstrip('\r\n')
        if len(self._file_stack) >= 10:
            print("Error: Includes nested too deeply",
                  file=sys.stderr)
            sys.exit(1)
        self._file_stack.append(
            (self._current_fp, self.linenum, self.current_ifname))
        try:
            new_fp = open(rest, "r")
        except OSError:
            print(f"Error: Could not read '{rest}'",
                  file=sys.stderr)
            sys.exit(1)
        self.current_ifname = rest
        self.linenum = 1
        self._current_fp = new_fp

    def _handle_redirect(self, stripped):
        """Handle @ redirect directive."""
        rest = stripped[1:].strip().rstrip('\r\n')
        if self.mbatching_flag:
            if self.outfp:
                self.outfp.close()
                self.outfp = None
            if rest:
                try:
                    self.outfp = open(rest, "w")
                except OSError:
                    print(f"Error: Could not write {rest}",
                          file=sys.stderr)
                    sys.exit(1)
        if self.listing_flag and rest:
            print(rest)
        self.linenum += 1
        yield Token(TokenType.NON_C_LINE, "", self.linenum - 1)

    def _handle_dollar_line(self, stripped):
        """Handle $ single-line C pass-through."""
        rest = stripped[1:]
        if self.outcfp:
            self.outcfp.write(rest)
        self.linenum += 1
        yield Token(TokenType.NON_C_LINE, "", self.linenum - 1)

    def _handle_hash_line(self, stripped):
        """Handle # C declaration line — tokenise."""
        body = stripped[1:].rstrip('\r\n')
        yield from self._tokenize_c_line(body, self.linenum)
        self.linenum += 1

    # ------------------------------------------------------------------
    # internal line-by-line driver
    # ------------------------------------------------------------------

    def _lex_stream(self):
        """Process all lines from self._current_fp, yielding tokens."""
        in_block_c = False

        while True:
            raw_line = self._current_fp.readline()
            if raw_line == "":
                # End of current file
                if self._file_stack:
                    self._current_fp.close()
                    self._current_fp, self.linenum, self.current_ifname = self._file_stack.pop()
                    continue
                else:
                    return       # real EOF

            # Strip the trailing newline for processing but track it
            line = raw_line

            # --- block C mode ($[ ... $]) ---
            if in_block_c:
                in_block_c = self._handle_block_c(line)
                continue

            # Determine line prefix — compute leading whitespace
            stripped = line.lstrip(' \t')
            leading_ws = line[:len(line) - len(stripped)]

            # In the original Flex lexer, leading [ \t] in INITIAL state is
            # always written to outfp regardless of what prefix follows.
            # We replicate this for all prefix types except pure text lines
            # (which include their own leading whitespace in the full line).
            if leading_ws and self.outfp and (
                    stripped.startswith("$") or
                    stripped.startswith("#") or
                    stripped.startswith("@") or
                    stripped.startswith("//")):
                self.outfp.write(leading_ws)

            # $[ block start
            if re.match(r'^\$\[[ \t\r]*\n?$', stripped):
                in_block_c = True
                self.linenum += 1
                continue

            if stripped.startswith("//"):
                yield from self._handle_comment()
                continue

            if stripped.startswith("@function"):
                yield from self._handle_function(stripped)
                continue

            if stripped.startswith("@include"):
                self._handle_include(stripped)
                continue

            if stripped.startswith("@"):
                yield from self._handle_redirect(stripped)
                continue

            if stripped.startswith("$") and not stripped.startswith("$["):
                yield from self._handle_dollar_line(stripped)
                continue

            if stripped.startswith("#"):
                yield from self._handle_hash_line(stripped)
                continue

            # Text line — copy to MATLAB output
            if self.outfp:
                self.outfp.write(line)
            self.linenum += 1
            yield Token(TokenType.NON_C_LINE, "", self.linenum - 1)

    # ------------------------------------------------------------------
    # tokenise a single '#' line body
    # ------------------------------------------------------------------

    def _tokenize_c_line(self, body, line):
        """Yield tokens for the body of a '#' line."""
        for m in _TOKEN_RE.finditer(body):
            comment, string, ident, number, punct = m.groups()
            if comment:
                break        # rest of line is a comment
            if string:
                yield Token(TokenType.STRING, string, line)
            elif ident:
                tt = KEYWORDS.get(ident, TokenType.ID)
                yield Token(tt, ident, line)
            elif number:
                yield Token(TokenType.NUMBER, number, line)
            elif punct:
                yield Token(TokenType.PUNCT, punct, line)



"""
mwrap_parser.py — Recursive descent parser for mwrap .mw files.

Copyright (c) 2007-2008  David Bindel
See the file COPYING for copying permissions

Converted to Python by Zydrunas Gimbutas (2026),
with assistance from Claude Code / Claude Opus 4.6 (Anthropic).
"""

import sys
from mwrap_ast import (
    Expr, TypeQual, Var, Func, VT,
    promote_int, id_string, add_inherits,
)
from mwrap_lexer import Lexer, Token, TokenType
from mwrap_typecheck import typecheck
from mwrap_mgen import print_matlab_call


class ParseError(Exception):
    pass


class Parser:
    """Recursive descent parser for mwrap '#' lines.

    Usage:
        parser = Parser(lexer, ctx, mexfunc=mexfunc)
        for tok in lexer.lex_file(filename):
            parser.feed(tok)
        # After all files:
        funcs = parser.funcs        # list[Func]
        err_flag = parser.err_flag
    """

    def __init__(self, lexer, ctx, mexfunc="mexfunction"):
        self.lexer = lexer
        self.ctx = ctx
        self.mexfunc = mexfunc

        # Function list
        self.funcs = []
        self.func_id = 0
        self.func_lookup = {}       # id_string -> first Func with that sig

        # Error tracking
        self.type_errs = 0
        self.err_flag = 0

        # Token buffer for current '#' line
        self._tokens = []
        self._pos = 0
        self._collecting = False    # True while inside a '#' line

    # ------------------------------------------------------------------
    # Token stream interface
    # ------------------------------------------------------------------

    def feed(self, tok):
        """Feed one token from the lexer."""
        if tok.type == TokenType.EOF:
            self._flush_pending(tok)
            return
        if tok.type == TokenType.NON_C_LINE:
            self._flush_pending(tok)
            return

        # Accumulate tokens from a '#' line
        self._tokens.append(tok)

        # A ';' terminates a statement — parse it
        if tok.type == TokenType.PUNCT and tok.value == ';':
            self._parse_line()
            self._tokens.clear()

    def _flush_pending(self, tok):
        """Report error if there are tokens pending without a ';'."""
        if self._tokens:
            tname = "NON_C_LINE" if tok.type == TokenType.NON_C_LINE else "$end"
            fname = self.lexer.current_ifname
            line = tok.line
            print(f"Parse error ({fname}:{line}): syntax error, unexpected {tname}, expecting ';'",
                  file=sys.stderr)
            self.err_flag += 1
            self._tokens.clear()

    def finish_file(self):
        """Call after all tokens from one input file have been fed."""
        # Flush any pending tokens (missing ';' at EOF)
        if self._tokens:
            fname = self.lexer.current_ifname
            line = self.lexer.linenum
            print(f"Parse error ({fname}:{line}): syntax error, unexpected $end, expecting ';'",
                  file=sys.stderr)
            self.err_flag += 1
            self._tokens.clear()
        if self.type_errs:
            print(f"{self.lexer.current_ifname}: {self.type_errs} type errors detected",
                  file=sys.stderr)
            self.err_flag += self.type_errs
        self.type_errs = 0

    # ------------------------------------------------------------------
    # Token access helpers
    # ------------------------------------------------------------------

    def _peek(self):
        if self._pos < len(self._tokens):
            return self._tokens[self._pos]
        return Token(TokenType.EOF, "", 0)

    def _advance(self):
        tok = self._peek()
        self._pos += 1
        return tok

    def _expect(self, ttype, value=None):
        tok = self._advance()
        if tok.type != ttype or (value is not None and tok.value != value):
            exp = f"{ttype.name}"
            if value:
                exp += f" '{value}'"
            self._error(f"Expected {exp}, got {tok.type.name} '{tok.value}'")
        return tok

    def _expect_punct(self, ch):
        return self._expect(TokenType.PUNCT, ch)

    def _at(self, ttype, value=None):
        t = self._peek()
        if t.type != ttype:
            return False
        if value is not None and t.value != value:
            return False
        return True

    def _at_punct(self, ch):
        return self._at(TokenType.PUNCT, ch)

    def _line(self):
        """Current line number for error messages."""
        if self._tokens:
            return self._tokens[0].line
        return self.lexer.linenum

    def _error(self, msg):
        fname = self.lexer.current_ifname
        line = self._line()
        print(f"Parse error ({fname}:{line}): {msg}", file=sys.stderr)
        raise ParseError(msg)

    # ------------------------------------------------------------------
    # Top-level statement parse
    # ------------------------------------------------------------------

    def _parse_line(self):
        """Parse one complete '#' line (already accumulated in _tokens)."""
        self._pos = 0
        try:
            self._statement()
        except ParseError:
            self.err_flag += 1

    def _statement(self):
        """statement ::= tdef | classdef | basevar '=' funcall | funcall"""
        tok = self._peek()

        if tok.type == TokenType.TYPEDEF:
            self._tdef()
            return

        if tok.type == TokenType.CLASS:
            self._classdef()
            return

        # Distinguish:  basevar = funcall   vs   funcall
        # A funcall starts with: ID -> ... | ID ( | FORTRAN ID | NEW ID
        # A basevar starts with: ID ID  or  ID qual ID
        # The disambiguator: look for '=' before '(' or ';'
        if self._has_assignment():
            bv = self._basevar()
            self._expect_punct('=')
            fc = self._funcall()
            fc.ret = [bv]
            self._finish_func(fc)
        else:
            fc = self._funcall()
            self._finish_func(fc)

    def _has_assignment(self):
        """Lookahead: is there a '=' before '(' or ';'?"""
        depth = 0
        for i in range(len(self._tokens)):
            t = self._tokens[i]
            if t.type == TokenType.PUNCT:
                if t.value == '[':
                    depth += 1
                elif t.value == ']':
                    depth -= 1
                elif depth == 0 and t.value == '=':
                    return True
                elif depth == 0 and t.value == '(':
                    return False
                elif depth == 0 and t.value == ';':
                    return False
        return False

    # ------------------------------------------------------------------
    # typedef / classdef
    # ------------------------------------------------------------------

    def _tdef(self):
        """TYPEDEF ID ID ';'"""
        self._expect(TokenType.TYPEDEF)
        space = self._expect(TokenType.ID).value
        name = self._expect(TokenType.ID).value
        self._expect_punct(';')

        if space == "numeric":
            self.ctx.add_scalar_type(name)
        elif space == "dcomplex":
            self.ctx.add_zscalar_type(name)
        elif space == "fcomplex":
            self.ctx.add_cscalar_type(name)
        elif space == "mxArray":
            self.ctx.add_mxarray_type(name)
        else:
            print(f"Unrecognized typespace: {space}", file=sys.stderr)
            self.type_errs += 1

    def _classdef(self):
        """CLASS ID ':' inheritslist ';'"""
        self._expect(TokenType.CLASS)
        child = self._expect(TokenType.ID).value
        self._expect_punct(':')
        parents = self._inheritslist()
        self._expect_punct(';')
        add_inherits(self.ctx, child, parents)

    def _inheritslist(self):
        """ID (',' ID)* — returns list[str]"""
        result = [self._expect(TokenType.ID).value]
        while self._at_punct(','):
            self._advance()
            result.append(self._expect(TokenType.ID).value)
        return result

    # ------------------------------------------------------------------
    # funcall / func
    # ------------------------------------------------------------------

    def _funcall(self):
        """funcall ::= func '(' args ')' ';'"""
        f = self._func()
        self._expect_punct('(')
        f.args = self._args()
        self._expect_punct(')')
        self._expect_punct(';')
        return f

    def _func(self):
        """func ::= ID '->' ID '.' ID | ID | FORTRAN ID | NEW ID"""
        tok = self._peek()

        if tok.type == TokenType.FORTRAN:
            self._advance()
            name = self._expect(TokenType.ID).value
            f = Func(None, None, name,
                     self.lexer.current_ifname, self._line())
            f.fort = True
            return f

        if tok.type == TokenType.NEW:
            self._advance()
            classname = self._expect(TokenType.ID).value
            return Func(None, classname, "new",
                        self.lexer.current_ifname, self._line())

        # ID — could be plain func, or  thisvar -> classname . method
        name = self._expect(TokenType.ID).value

        if self._at_punct('-'):
            # thisvar -> classname . method
            self._advance()  # '-'
            self._expect_punct('>')
            classname = self._expect(TokenType.ID).value
            self._expect_punct('.')
            method = self._expect(TokenType.ID).value
            return Func(name, classname, method,
                        self.lexer.current_ifname, self._line())

        return Func(None, None, name,
                    self.lexer.current_ifname, self._line())

    # ------------------------------------------------------------------
    # args / var
    # ------------------------------------------------------------------

    def _args(self):
        """args ::= var (',' var)* | ε — returns list[Var]"""
        if self._at_punct(')'):
            return []
        result = [self._var()]
        while self._at_punct(','):
            self._advance()
            result.append(self._var())
        return result

    def _var(self):
        """var ::= [devicespec] [iospec] TYPE [qual] (NAME | NUMBER | STRING)
           OR:  [devicespec] [iospec] TYPE NAME [aqual]    (post-name array)
        """
        devicespec = self._devicespec()
        iospec = self._iospec()
        basetype = promote_int(self.ctx, self._expect(TokenType.ID).value)

        # Now we may see:
        #   quals then NAME/NUMBER/STRING
        #   NAME/NUMBER/STRING then optional aqual
        #   Just NAME/NUMBER/STRING (no qual)
        tok = self._peek()

        if tok.type == TokenType.PUNCT and tok.value in ('*', '&', '['):
            # quals before name
            qual = self._quals()
            name = self._name_or_literal()
            return Var(devicespec, iospec, basetype, qual, name)

        # NAME/NUMBER/STRING first, then optional aqual
        name = self._name_or_literal()

        # Check for post-name aqual:  name [ ... ] or name [ ... ] &
        if self._at_punct('['):
            qual = self._aqual()
            return Var(devicespec, iospec, basetype, qual, name)

        return Var(devicespec, iospec, basetype, None, name)

    def _name_or_literal(self):
        tok = self._peek()
        if tok.type in (TokenType.ID, TokenType.NUMBER, TokenType.STRING):
            self._advance()
            return tok.value
        self._error(f"Expected name, number or string, got {tok.type.name} '{tok.value}'")

    def _devicespec(self):
        tok = self._peek()
        if tok.type == TokenType.CPU:
            self._advance()
            return 'c'
        if tok.type == TokenType.GPU:
            self._advance()
            return 'g'
        return 'c'

    def _iospec(self):
        tok = self._peek()
        m = {
            TokenType.INPUT:   'i',
            TokenType.OUTPUT:  'o',
            TokenType.INOUT:   'b',
        }
        if tok.type in m:
            self._advance()
            return m[tok.type]
        return 'i'

    def _quals(self):
        """quals ::= '*' | '&' | aqual"""
        if self._at_punct('*'):
            self._advance()
            return TypeQual('*')
        if self._at_punct('&'):
            self._advance()
            return TypeQual('&')
        return self._aqual()

    def _aqual(self):
        """aqual ::= arrayspec ['&']"""
        exprs = self._arrayspec()
        if self._at_punct('&'):
            self._advance()
            return TypeQual('r', exprs)
        return TypeQual('a', exprs)

    def _arrayspec(self):
        """'[' exprs ']' — returns list[Expr]"""
        self._expect_punct('[')
        e = self._exprs()
        self._expect_punct(']')
        return e

    def _exprs(self):
        """exprs ::= expr (',' expr)* | ε — returns list[Expr]"""
        if self._at_punct(']'):
            return []
        result = [self._expr()]
        while self._at_punct(','):
            self._advance()
            result.append(self._expr())
        return result

    def _expr(self):
        """expr ::= ID | NUMBER"""
        tok = self._peek()
        if tok.type in (TokenType.ID, TokenType.NUMBER):
            self._advance()
            return Expr(tok.value)
        self._error(f"Expected expression, got {tok.type.name} '{tok.value}'")

    def _basevar(self):
        """basevar ::= ID [quals] ID  — always output, cpu"""
        basetype = promote_int(self.ctx, self._expect(TokenType.ID).value)

        # Peek: quals or name?
        tok = self._peek()
        if tok.type == TokenType.PUNCT and tok.value in ('*', '&', '['):
            qual = self._quals()
            name = self._expect(TokenType.ID).value
            return Var('c', 'o', basetype, qual, name)

        # NAME then optional aqual
        name = self._expect(TokenType.ID).value

        if self._at_punct('['):
            qual = self._aqual()
            return Var('c', 'o', basetype, qual, name)

        return Var('c', 'o', basetype, None, name)

    # ------------------------------------------------------------------
    # Post-parse: typecheck, MATLAB stub, add to func list
    # ------------------------------------------------------------------

    def _finish_func(self, func):
        """Typecheck, emit MATLAB stub, add to function list."""
        self.func_id += 1
        func.id = self.func_id

        self.type_errs += typecheck(self.ctx, func, self._line())

        if self.lexer.outfp:
            print_matlab_call(self.lexer.outfp, self.ctx, func, self.mexfunc)

        self._add_func(func)

    def _add_func(self, func):
        """Add func to list; deduplicate via id_string."""
        ids = id_string(self.ctx, func)

        first = self.func_lookup.get(ids)
        if first:
            # Same signature — add to first occurrence's same list
            first.same.append(func)
        else:
            # New unique signature — append to main list
            self.funcs.append(func)
            self.func_lookup[ids] = func



"""
mwrap_ast.py — AST nodes, type registries, and helpers for mwrap.

Copyright (c) 2007-2008  David Bindel
See the file COPYING for copying permissions

Converted to Python by Zydrunas Gimbutas (2026),
with assistance from Claude Code / Claude Opus 4.6 (Anthropic).
"""

from enum import IntEnum
from dataclasses import dataclass, field
from typing import Optional


# ---------------------------------------------------------------------------
# VT_* classification codes
# ---------------------------------------------------------------------------

class VT(IntEnum):
    unk       = 0
    obj       = 1
    array     = 2
    carray    = 3
    zarray    = 4
    rarray    = 5
    scalar    = 6
    cscalar   = 7
    zscalar   = 8
    string    = 9
    mx        = 10
    p_obj     = 11
    p_scalar  = 12
    p_cscalar = 13
    p_zscalar = 14
    r_obj     = 15
    r_scalar  = 16
    r_cscalar = 17
    r_zscalar = 18
    const     = 19


# ---------------------------------------------------------------------------
# iospec / VT classification helpers (single source of truth)
# ---------------------------------------------------------------------------

def iospec_is_input(c):   return c in ('i', 'b')
def iospec_is_output(c):  return c in ('o', 'b')
def iospec_is_inonly(c):  return c == 'i'

def is_array(tinfo):
    return tinfo in (VT.array, VT.carray, VT.zarray)

def is_obj(tinfo):
    return tinfo in (VT.obj, VT.p_obj, VT.r_obj)

def complex_tinfo(v):
    return v.tinfo in (VT.carray, VT.zarray,
                       VT.cscalar, VT.zscalar,
                       VT.r_cscalar, VT.r_zscalar,
                       VT.p_cscalar, VT.p_zscalar)

def nullable_return(f):
    return (f.ret and f.ret[0].tinfo in (
        VT.string, VT.array, VT.carray, VT.zarray,
        VT.p_scalar, VT.p_cscalar, VT.p_zscalar))


# ---------------------------------------------------------------------------
# AST node dataclasses
# ---------------------------------------------------------------------------

@dataclass
class Expr:
    value: str
    input_label: int = -1


@dataclass
class TypeQual:
    qual: str          # '*', '&', 'a' (array), 'r' (array ref)
    args: list = field(default_factory=list)  # list[Expr]


@dataclass
class Var:
    devicespec: str    # 'c' (cpu) or 'g' (gpu)
    iospec: str        # 'i','o','b'
    basetype: str
    qual: Optional[TypeQual]
    name: str
    tinfo: int = VT.unk
    input_label: int = -1
    output_label: int = -1


@dataclass
class Func:
    thisv: Optional[str]
    classv: Optional[str]
    funcv: str
    fname: str
    line: int
    fort: bool = False
    id: int = -1
    args: list = field(default_factory=list)  # list[Var]
    ret: list = field(default_factory=list)   # list[Var] (0 or 1 elements)
    same: list = field(default_factory=list)  # list[Func] — duplicate signatures


# ---------------------------------------------------------------------------
# MwrapContext — all mutable state in one object
# ---------------------------------------------------------------------------

class MwrapContext:
    def __init__(self):
        # CLI flags
        self.mw_use_gpu = False
        self.mw_generate_catch = False
        self.mw_use_c99_complex = False
        self.mw_use_cpp_complex = False
        self.mw_promote_int = 0

        # Type registries
        self.scalar_decls = set()
        self.cscalar_decls = set()
        self.zscalar_decls = set()
        self.mxarray_decls = set()
        self.class_decls = {}   # str -> list[str]

        # Type usage tracking
        self.mw_use_int32_t = 0
        self.mw_use_int64_t = 0
        self.mw_use_uint32_t = 0
        self.mw_use_uint64_t = 0
        self.mw_use_longlong = 0
        self.mw_use_ulonglong = 0
        self.mw_use_ulong = 0
        self.mw_use_uint = 0
        self.mw_use_ushort = 0
        self.mw_use_uchar = 0

    def init_scalar_types(self):
        self.scalar_decls.clear()
        for t in ("double", "float",
                  "long", "int", "short", "char",
                  "ulong", "uint", "ushort", "uchar",
                  "int32_t", "int64_t", "uint32_t", "uint64_t",
                  "bool", "size_t", "ptrdiff_t"):
            self.scalar_decls.add(t)

    def is_scalar_type(self, name):    return name in self.scalar_decls
    def is_cscalar_type(self, name):   return name in self.cscalar_decls
    def is_zscalar_type(self, name):   return name in self.zscalar_decls
    def is_mxarray_type(self, name):   return name in self.mxarray_decls

    def add_scalar_type(self, name):   self.scalar_decls.add(name)
    def add_cscalar_type(self, name):  self.cscalar_decls.add(name)
    def add_zscalar_type(self, name):  self.zscalar_decls.add(name)
    def add_mxarray_type(self, name):  self.mxarray_decls.add(name)


# ---------------------------------------------------------------------------
# Functions that need context
# ---------------------------------------------------------------------------

def promote_int(ctx, name: str) -> str:
    if name == "int32_t":  ctx.mw_use_int32_t = 1
    if name == "int64_t":  ctx.mw_use_int64_t = 1
    if name == "uint32_t": ctx.mw_use_uint32_t = 1
    if name == "uint64_t": ctx.mw_use_uint64_t = 1
    if name == "ulong":    ctx.mw_use_ulong = 1
    if name == "uint":     ctx.mw_use_uint = 1
    if name == "ushort":   ctx.mw_use_ushort = 1
    if name == "uchar":    ctx.mw_use_uchar = 1

    if ctx.mw_promote_int == 1:
        if name == "uint":  ctx.mw_use_ulong = 1
        if name == "int":   return "long"
        if name == "uint":  return "ulong"
    elif ctx.mw_promote_int == 2:
        if name == "uint":  ctx.mw_use_ulong = 1
        if name == "ulong": ctx.mw_use_ulong = 1
        if name == "int":   return "long"
        if name == "long":  return "long"
        if name == "uint":  return "ulong"
        if name == "ulong": return "ulong"
    elif ctx.mw_promote_int == 3:
        if name == "int":   ctx.mw_use_int32_t = 1
        if name == "long":  ctx.mw_use_int64_t = 1
        if name == "uint":  ctx.mw_use_uint32_t = 1
        if name == "ulong": ctx.mw_use_uint64_t = 1
        if name == "int":   return "int32_t"
        if name == "long":  return "int32_t"
        if name == "uint":  return "uint64_t"
        if name == "ulong": return "uint64_t"
    elif ctx.mw_promote_int == 4:
        if name == "int":   ctx.mw_use_int64_t = 1
        if name == "long":  ctx.mw_use_int64_t = 1
        if name == "uint":  ctx.mw_use_uint64_t = 1
        if name == "ulong": ctx.mw_use_uint64_t = 1
        if name == "int":   return "int64_t"
        if name == "long":  return "int64_t"
        if name == "uint":  return "uint64_t"
        if name == "ulong": return "uint64_t"
    return name


def add_inherits(ctx, childname: str, parents: list):
    """Register childname as child of each parent in parents list."""
    for parent in parents:
        if parent not in ctx.class_decls:
            ctx.class_decls[parent] = []
        ctx.class_decls[parent].insert(0, childname)


# ---------------------------------------------------------------------------
# ID string (canonical signature for deduplication)
# ---------------------------------------------------------------------------

def _id_expr(args: list) -> str:
    return "x" * len(args)


def _id_qual(q: Optional[TypeQual]) -> str:
    if not q:
        return ""
    if q.qual == 'a':
        return "[" + _id_expr(q.args) + "]"
    return q.qual


def _id_var_single(ctx, v: Var) -> str:
    name = ""
    if v.devicespec == 'c':
        name += "c "
    elif v.devicespec == 'g':
        name += "g "

    io = v.iospec
    if io == 'i':   name += "i "
    elif io == 'o': name += "o "
    else:           name += "io "

    name += promote_int(ctx, v.basetype)
    name += _id_qual(v.qual)
    if v.tinfo == VT.const:
        name += " " + v.name
    return name


def _id_var(ctx, vars: list) -> str:
    return ", ".join(_id_var_single(ctx, v) for v in vars)


def id_string(ctx, f) -> str:
    if not f:
        return ""
    name = ""
    if f.ret:
        name += _id_var(ctx, f.ret) + " = "
    if f.thisv:
        name += f.thisv + "->" + f.classv + "."
    name += f.funcv + "(" + _id_var(ctx, f.args) + ")"
    return name


# ---------------------------------------------------------------------------
# AST printing (for C comments)
# ---------------------------------------------------------------------------

def _print_expr(args: list) -> str:
    return ", ".join(e.value for e in args)


def _print_qual(q):
    if not q:
        return ""
    if q.qual == 'a':
        return "[" + _print_expr(q.args) + "]"
    return q.qual


def _print_devicespec(v):
    if v.devicespec == 'g':
        return "gpu "
    return ""


def _print_iospec(v):
    m = {'o': "output ", 'b': "inout "}
    return m.get(v.iospec, "")


def _print_var(v):
    return v.basetype + _print_qual(v.qual) + " " + v.name


def _print_args(vars: list) -> str:
    return ", ".join(
        _print_devicespec(v) + _print_iospec(v) + _print_var(v)
        for v in vars
    )


def print_func(f):
    """Human-readable translation of Func AST (for C comments)."""
    if not f:
        return ""
    s = ""
    if f.ret:
        s += _print_var(f.ret[0]) + " = "
    if f.thisv:
        s += f"{f.thisv}->{f.classv}."
    s += f"{f.funcv}({_print_args(f.args)});\n"
    return s



"""
mwrap_cgen.py — C/MEX code generator.

Copyright (c) 2007-2008  David Bindel
See the file COPYING for copying permissions

Converted to Python by Zydrunas Gimbutas (2026),
with assistance from Claude Code / Claude Opus 4.6 (Anthropic).
"""

import sys
from dataclasses import dataclass
from mwrap_ast import (
    VT, Expr, TypeQual, Var, Func,
    id_string, print_func,
    is_array, is_obj, complex_tinfo, nullable_return,
)


# ===================================================================
# Type property table — single source of truth for type dispatch
# ===================================================================

@dataclass(frozen=True)
class TypeProps:
    mxclass: str             # "mxDOUBLE_CLASS", "mxSINGLE_CLASS", etc.
    accessor: str | None     # interleaved API accessor: "mxGetDoubles", etc.
    is_single: bool          # True → use _single copier variants
    scalar_getter: str       # "mxWrapGetScalar" / "_single" / "_char"
    scalar_class: str        # mxClass for scalar validation
    direct_input: bool = False  # True → use direct accessor for input-only arrays

_DEFAULT_PROPS = TypeProps("mxVOID_CLASS", None, False, "mxWrapGetScalar", "mxDOUBLE_CLASS")

TYPE_PROPS = {
    "double":   TypeProps("mxDOUBLE_CLASS", "mxGetDoubles",        False, "mxWrapGetScalar",        "mxDOUBLE_CLASS", True),
    "float":    TypeProps("mxSINGLE_CLASS", "mxGetSingles",        True,  "mxWrapGetScalar_single", "mxSINGLE_CLASS", True),
    "int32_t":  TypeProps("mxINT32_CLASS",  "mxGetInt32s",         False, "mxWrapGetScalar",        "mxDOUBLE_CLASS"),
    "int64_t":  TypeProps("mxINT64_CLASS",  "mxGetInt64s",         False, "mxWrapGetScalar",        "mxDOUBLE_CLASS"),
    "uint32_t": TypeProps("mxUINT32_CLASS", "mxGetUint32s",        False, "mxWrapGetScalar",        "mxDOUBLE_CLASS"),
    "uint64_t": TypeProps("mxUINT64_CLASS", "mxGetUint64s",        False, "mxWrapGetScalar",        "mxDOUBLE_CLASS"),
    "dcomplex": TypeProps("mxDOUBLE_CLASS", "mxGetComplexDoubles", False, "mxWrapGetScalar",        "mxDOUBLE_CLASS"),
    "fcomplex": TypeProps("mxSINGLE_CLASS", "mxGetComplexSingles", True,  "mxWrapGetScalar_single", "mxSINGLE_CLASS"),
    "char":     TypeProps("mxCHAR_CLASS",   None,                  False, "mxWrapGetScalar_char",   "mxCHAR_CLASS"),
}


def _type_props(name):
    return TYPE_PROPS.get(name, _DEFAULT_PROPS)


def _copier_suffix(bt):
    """Return 'single_' for float-precision types, '' otherwise."""
    return "single_" if _type_props(bt).is_single else ""


# ===================================================================
# Utility functions
# ===================================================================

def basetype_to_cucomplex(name):
    if name == "fcomplex": return "cuFloatComplex"
    if name == "dcomplex": return "cuDoubleComplex"
    return name


def basetype_to_mxclassid(name):
    return _type_props(name).mxclass


def vname(v):
    if v.iospec == 'o':
        return f"out{v.output_label}_"
    return f"in{v.input_label}_"


def has_fortran(funcs):
    return any(f.fort for f in funcs)


def max_routine_id(funcs):
    maxid = 0
    for f in funcs:
        if f.id > maxid:
            maxid = f.id
    return maxid


def _alloc_size_expr(args):
    """Return C expression for product of dim args."""
    if not args:
        return "1"
    return "*".join(f"dim{e.input_label}_" for e in args)


def _interleaved_branch(fp, interleaved, fallback):
    """Emit #if MX_HAS_INTERLEAVED_COMPLEX / #else / #endif block."""
    fp.write(f"#if MX_HAS_INTERLEAVED_COMPLEX\n{interleaved}#else\n{fallback}#endif\n")


# ===================================================================
# Complex type definitions
# ===================================================================

def mex_cpp_complex(fp):
    fp.write("#include <complex>\n\n"
           "typedef std::complex<double> dcomplex;\n"
           "#define real_dcomplex(z) std::real(z)\n"
           "#define imag_dcomplex(z) std::imag(z)\n"
           "#define setz_dcomplex(z,r,i)  *z = dcomplex(r,i)\n\n"
           "typedef std::complex<float> fcomplex;\n"
           "#define real_fcomplex(z) std::real(z)\n"
           "#define imag_fcomplex(z) std::imag(z)\n"
           "#define setz_fcomplex(z,r,i)  *z = fcomplex(r,i)\n\n")


def mex_c99_complex(fp):
    fp.write("#include <complex.h>\n\n"
           "typedef _Complex double dcomplex;\n"
           "#define real_dcomplex(z) creal(z)\n"
           "#define imag_dcomplex(z) cimag(z)\n"
           "#define setz_dcomplex(z,r,i)  *z = r + i*_Complex_I\n\n"
           "typedef _Complex float fcomplex;\n"
           "#define real_fcomplex(z) crealf(z)\n"
           "#define imag_fcomplex(z) cimagf(z)\n"
           "#define setz_fcomplex(z,r,i)  *z = r + i*_Complex_I\n\n")


def mex_gpucpp_complex(fp):
    fp.write("#include <cuComplex.h>\n\n")


# ===================================================================
# Copier instantiation
# ===================================================================

def _mex_define_copiers_type(fp, ctx, name):
    """Emit copier macro calls for one scalar type."""
    # Skip types not actually used
    if name == "int32_t"  and not ctx.mw_use_int32_t:  return
    if name == "int64_t"  and not ctx.mw_use_int64_t:  return
    if name == "uint32_t" and not ctx.mw_use_uint32_t: return
    if name == "uint64_t" and not ctx.mw_use_uint64_t: return
    if name == "ulong"    and not ctx.mw_use_ulong:    return
    if name == "uint"     and not ctx.mw_use_uint:     return
    if name == "ushort"   and not ctx.mw_use_ushort:   return
    if name == "uchar"    and not ctx.mw_use_uchar:    return

    fp.write(f"mxWrapGetArrayDef(mxWrapGetArray_{name}, {name})\n")
    fp.write(f"mxWrapCopyDef    (mxWrapCopy_{name},     {name})\n")
    fp.write(f"mxWrapReturnDef  (mxWrapReturn_{name},   {name})\n")
    fp.write(f"mxWrapGetArrayDef_single(mxWrapGetArray_single_{name}, {name})\n")
    fp.write(f"mxWrapCopyDef_single    (mxWrapCopy_single_{name},     {name})\n")
    fp.write(f"mxWrapReturnDef_single  (mxWrapReturn_single_{name},   {name})\n")


def _mex_define_zcopiers(fp, name, ztype):
    """Emit complex copier macro calls for one complex type."""
    fp.write(f"mxWrapGetScalarZDef(mxWrapGetScalar_{name}, {name},\n"
           f"                    {ztype}, setz_{name})\n")
    fp.write(f"mxWrapGetArrayZDef (mxWrapGetArray_{name}, {name},\n"
           f"                    {ztype}, setz_{name})\n")
    fp.write(f"mxWrapCopyZDef     (mxWrapCopy_{name}, {name},\n"
           f"                    real_{name}, imag_{name})\n")
    fp.write(f"mxWrapReturnZDef   (mxWrapReturn_{name}, {name},\n"
           f"                    real_{name}, imag_{name})\n")
    fp.write(f"mxWrapGetScalarZDef_single(mxWrapGetScalar_single_{name}, {name},\n"
           f"                    {ztype}, setz_{name})\n")
    fp.write(f"mxWrapGetArrayZDef_single (mxWrapGetArray_single_{name}, {name},\n"
           f"                    {ztype}, setz_{name})\n")
    fp.write(f"mxWrapCopyZDef_single     (mxWrapCopy_single_{name}, {name},\n"
           f"                    real_{name}, imag_{name})\n")
    fp.write(f"mxWrapReturnZDef_single   (mxWrapReturn_single_{name}, {name},\n"
           f"                    real_{name}, imag_{name})\n")


def mex_define_copiers(fp, ctx):
    fp.write("\n\n\n/* Array copier definitions */\n")
    for name in sorted(ctx.scalar_decls):
        _mex_define_copiers_type(fp, ctx, name)
    for name in sorted(ctx.cscalar_decls):
        _mex_define_zcopiers(fp, name, "float")
    for name in sorted(ctx.zscalar_decls):
        _mex_define_zcopiers(fp, name, "double")
    fp.write("\n")


# ===================================================================
# Fortran name mangling
# ===================================================================

def _fortran_funcs(funcs):
    """Yield unique Func objects for fortran functions."""
    seen = set()
    for f in funcs:
        if f.fort and f.funcv not in seen:
            seen.add(f.funcv)
            yield f


def mex_define_fnames(fp, funcs):
    fp.write("#if defined(MWF77_CAPS)\n")
    for fc in _fortran_funcs(funcs):
        fp.write(f"#define MWF77_{fc.funcv} {fc.funcv.upper()}\n")
    fp.write("#elif defined(MWF77_UNDERSCORE1)\n")
    for fc in _fortran_funcs(funcs):
        fp.write(f"#define MWF77_{fc.funcv} {fc.funcv.lower()}_\n")
    fp.write("#elif defined(MWF77_UNDERSCORE0)\n")
    for fc in _fortran_funcs(funcs):
        fp.write(f"#define MWF77_{fc.funcv} {fc.funcv.lower()}\n")
    fp.write("#else /* f2c convention */\n")
    for fc in _fortran_funcs(funcs):
        low = fc.funcv.lower()
        suffix = "__" if '_' in low else "_"
        fp.write(f"#define MWF77_{fc.funcv} {low}{suffix}\n")
    fp.write("#endif\n\n")


def _mex_fortran_arg(fp, args):
    parts = []
    for v in args:
        if v.tinfo == VT.mx:
            parts.append("const mxArray*")
        else:
            parts.append(f"{v.basetype}*")
    fp.write(", ".join(parts))


def mex_fortran_decls(fp, funcs):
    fp.write("#ifdef __cplusplus\n"
           "extern \"C\" { /* Prevent C++ name mangling */\n"
           "#endif\n\n"
           "#ifndef MWF77_RETURN\n"
           "#define MWF77_RETURN int\n"
           "#endif\n\n")
    for fc in _fortran_funcs(funcs):
        if fc.ret:
            fp.write(f"{fc.ret[0].basetype} ")
        else:
            fp.write("MWF77_RETURN ")
        fp.write(f"MWF77_{fc.funcv}(")
        _mex_fortran_arg(fp, fc.args)
        fp.write(");\n")
    fp.write("\n#ifdef __cplusplus\n"
           "} /* end extern C */\n"
           "#endif\n\n")


# ===================================================================
# Class polymorphism getters
# ===================================================================

def _mex_casting_getter_type(fp, name):
    fp.write(f"    {name}* p_{name} = NULL;\n"
           f"    sscanf(pbuf, \"{name}:%p\", &p_{name});\n"
           f"    if (p_{name})\n"
           f"        return p_{name};\n\n")


def _mex_casting_getter(fp, cname, inherits):
    fp.write(f"\n{cname}* mxWrapGetP_{cname}(const mxArray* a, const char** e)\n")
    fp.write("{\n"
           "    char pbuf[128];\n"
           "    if (mxGetClassID(a) == mxDOUBLE_CLASS &&\n"
           "        mxGetM(a)*mxGetN(a) == 1 &&\n"
           "#if MX_HAS_INTERLEAVED_COMPLEX\n"
           "        ((mxIsComplex(a) ? ((*mxGetComplexDoubles(a)).real == 0 && (*mxGetComplexDoubles(a)).imag == 0) : *mxGetDoubles(a) == 0))\n"
           "#else\n"
           "        *mxGetPr(a) == 0\n"
           "#endif\n"
           "        )\n"
           "        return NULL;\n"
           "    if (!mxIsChar(a)) {\n"
           "#ifdef R2008OO\n"
           f"        mxArray* ap = mxGetProperty(a, 0, \"mwptr\");\n"
           f"        if (ap)\n"
           f"            return mxWrapGetP_{cname}(ap, e);\n"
           "#endif\n"
           "        *e = \"Invalid pointer\";\n"
           "        return NULL;\n"
           "    }\n"
           "    mxGetString(a, pbuf, sizeof(pbuf));\n\n")
    _mex_casting_getter_type(fp, cname)
    for name in inherits:
        _mex_casting_getter_type(fp, name)
    fp.write(f"    *e = \"Invalid pointer to {cname}\";\n"
           f"    return NULL;\n"
           f"}}\n\n")


def mex_casting_getters(fp, ctx):
    for parent in sorted(ctx.class_decls.keys()):
        _mex_casting_getter(fp, parent, ctx.class_decls[parent])


# ===================================================================
# Per-stub helpers: declarations, unpack, check, alloc, call, marshal, dealloc
# ===================================================================

def _declare_type(v):
    """Return C type string for a variable declaration."""
    if is_obj(v.tinfo) or is_array(v.tinfo):
        if v.devicespec == 'g':
            return f"{basetype_to_cucomplex(v.basetype)}*"
        return f"{v.basetype}*"
    if v.tinfo == VT.rarray:
        if v.devicespec == 'g':
            return f"const {basetype_to_cucomplex(v.basetype)}*"
        return f"const {v.basetype}*"
    if v.tinfo in (VT.scalar, VT.cscalar, VT.zscalar,
                   VT.r_scalar, VT.r_cscalar, VT.r_zscalar,
                   VT.p_scalar, VT.p_cscalar, VT.p_zscalar):
        return v.basetype
    if v.tinfo == VT.string:
        return "char*"
    if v.tinfo == VT.mx:
        if v.iospec == 'i':
            return "const mxArray*"
        return "mxArray*"
    assert False, f"Unknown tinfo {v.tinfo} for {v.name}"


# --- Step 1: Declare locals ---

def _declare_in_args(fp, args):
    for v in args:
        if v.iospec != 'o' and v.tinfo != VT.const:
            tb = _declare_type(v)
            if is_array(v.tinfo) or is_obj(v.tinfo) or v.tinfo == VT.string:
                fp.write(f"    {tb:10s}  in{v.input_label}_ =0; /* {v.name:10s} */\n")
                if v.devicespec == 'g':
                    fp.write(f"    {'mxGPUArray const':10s} *mxGPUArray_in{v.input_label}_ =0; /* {v.name:10s} */\n")
            else:
                fp.write(f"    {tb:10s}  in{v.input_label}_;    /* {v.name:10s} */\n")


def _declare_out_args(fp, args):
    for v in args:
        if v.iospec == 'o' and v.tinfo != VT.mx:
            tb = _declare_type(v)
            if is_array(v.tinfo) or is_obj(v.tinfo) or v.tinfo == VT.string:
                fp.write(f"    {tb:10s}  out{v.output_label}_=0; /* {v.name:10s} */\n")
                if v.devicespec == 'g':
                    fp.write(f"    {'mxGPUArray':10s} *mxGPUArray_out{v.output_label}_ =0; /* {v.name:10s} */\n")
                    fp.write(f"    {'mwSize':10s} gpu_outdims{v.output_label}_[2] = {{0,0}}; /* {v.name:10s} dims*/\n")
            else:
                fp.write(f"    {tb:10s}  out{v.output_label}_;   /* {v.name:10s} */\n")


def _declare_dim_args_expr(fp, args):
    for e in args:
        fp.write(f"    {'mwSize':10s}  dim{e.input_label}_;   /* {e.value:10s} */\n")


def _declare_dim_args_var(fp, vars):
    for v in vars:
        if v.qual:
            _declare_dim_args_expr(fp, v.qual.args)


def _declare_args(fp, f):
    if f.thisv:
        tb = f"{f.classv}*"
        fp.write(f"    {tb:10s}  in0_ =0; /* {f.thisv:10s} */\n")
    _declare_in_args(fp, f.args)
    if not nullable_return(f):
        _declare_out_args(fp, f.ret)
    _declare_out_args(fp, f.args)
    _declare_dim_args_var(fp, f.ret)
    _declare_dim_args_var(fp, f.args)
    if f.ret or f.args or f.thisv:
        fp.write("\n")


# --- Step 2: Unpack dims ---

def _unpack_dims_expr(fp, args):
    count = 0
    for e in args:
        fp.write(f"    dim{e.input_label}_ = (mwSize) mxWrapGetScalar(prhs[{e.input_label}], &mw_err_txt_);\n")
        count += 1
    return count


def _unpack_dims_var(fp, vars):
    count = 0
    for v in vars:
        if v.qual:
            count += _unpack_dims_expr(fp, v.qual.args)
    return count


def _unpack_dims(fp, f):
    c = _unpack_dims_var(fp, f.ret) + _unpack_dims_var(fp, f.args)
    if c:
        fp.write("\n")


# --- Step 3: Check dim consistency ---

def _check_dims(fp, args):
    for v in args:
        if (v.iospec != 'o' and is_array(v.tinfo) and
                v.qual and v.qual.args and v.devicespec != 'g'):
            a = v.qual.args
            if len(a) > 1:
                fp.write(f"    if (mxGetM(prhs[{v.input_label}]) != dim{a[0].input_label}_ ||\n"
                       f"        mxGetN(prhs[{v.input_label}]) != dim{a[1].input_label}_) {{\n"
                       f"        mw_err_txt_ = \"Bad argument size: {v.name}\";\n"
                       f"        goto mw_err_label;\n"
                       f"    }}\n\n")
            else:
                fp.write(f"    if (mxGetM(prhs[{v.input_label}])*mxGetN(prhs[{v.input_label}]) != dim{a[0].input_label}_) {{\n"
                       f"        mw_err_txt_ = \"Bad argument size: {v.name}\";"
                       f"        goto mw_err_label;\n"
                       f"    }}\n\n")


# --- Step 4: Unpack inputs ---

def _cast_get_p(fp, ctx, basetype, input_label):
    fp.write(f"    in{input_label}_ = ")
    if ctx.is_mxarray_type(basetype):
        fp.write(f"mxWrapGet_{basetype}(prhs[{input_label}], &mw_err_txt_);\n")
    elif basetype not in ctx.class_decls:
        fp.write(f"({basetype}*) mxWrapGetP(prhs[{input_label}], \"{basetype}:%p\", &mw_err_txt_);\n")
    else:
        fp.write(f"mxWrapGetP_{basetype}(prhs[{input_label}], &mw_err_txt_);\n")
    fp.write("    if (mw_err_txt_)\n"
           "        goto mw_err_label;\n\n")


def _unpack_input_array(fp, v):
    il = v.input_label
    bt = v.basetype

    # --- Regular (copy) path for CPU ---
    if v.devicespec != 'g':
        tp = _type_props(bt)
        cs = _copier_suffix(bt)
        fp.write(f"    if (mxGetM(prhs[{il}])*mxGetN(prhs[{il}]) != 0) {{\n")
        if complex_tinfo(v) and bt in TYPE_PROPS:
            # Known complex types: class check + copier
            fp.write(f"        if( mxGetClassID(prhs[{il}]) != {tp.mxclass} )\n"
                   f"            mw_err_txt_ = \"Invalid array argument, {tp.mxclass} expected\";\n"
                   f"        if (mw_err_txt_) goto mw_err_label;\n"
                   f"        in{il}_ = mxWrapGetArray_{cs}{bt}(prhs[{il}], &mw_err_txt_);\n"
                   f"        if (mw_err_txt_)\n"
                   f"            goto mw_err_label;\n")
        elif tp.direct_input and v.iospec == 'i':
            # float/double input-only: class check + direct accessor
            fp.write(f"        if( mxGetClassID(prhs[{il}]) != {tp.mxclass} )\n"
                   f"            mw_err_txt_ = \"Invalid array argument, {tp.mxclass} expected\";\n"
                   f"        if (mw_err_txt_) goto mw_err_label;\n"
                   f"#if MX_HAS_INTERLEAVED_COMPLEX\n"
                   f"        in{il}_ = {tp.accessor}(prhs[{il}]);\n"
                   f"#else\n")
            if bt == "double":
                fp.write(f"        in{il}_ = mxGetPr(prhs[{il}]);\n")
            else:
                fp.write(f"        in{il}_ = ({bt}*) mxGetData(prhs[{il}]);\n")
            fp.write(f"#endif\n")
        else:
            # All other types (including float/double inout): copier
            fp.write(f"        in{il}_ = mxWrapGetArray_{cs}{bt}(prhs[{il}], &mw_err_txt_);\n"
                   f"        if (mw_err_txt_)\n"
                   f"            goto mw_err_label;\n")
        fp.write(f"    }} else\n"
               f"        in{il}_ = NULL;\n")
        fp.write("\n")

    # --- GPU path ---
    if v.devicespec == 'g':
        if v.iospec in ('i', 'b'):
            cutype = basetype_to_cucomplex(bt)
            fp.write(f"    // extract input GPU array pointer\n"
                   f"    if(!(mxIsGPUArray(prhs[{il}])))\n"
                   f"        mw_err_txt_ = \"Invalid array argument, gpuArray expected\";\n"
                   f"    if (mw_err_txt_) goto mw_err_label;\n"
                   f"    mxGPUArray_in{il}_ = mxGPUCreateFromMxArray(prhs[{il}]);\n"
                   f"    in{il}_ = ({cutype} *)mxGPUGetDataReadOnly(mxGPUArray_in{il}_);\n\n")


def _unpack_input_string(fp, v):
    il = v.input_label
    if not (v.qual and v.qual.args):
        fp.write(f"    in{il}_ = mxWrapGetString(prhs[{il}], &mw_err_txt_);\n"
               f"    if (mw_err_txt_)\n"
               f"        goto mw_err_label;\n")
    else:
        sz = _alloc_size_expr(v.qual.args)
        fp.write(f"    in{il}_ = (char*) mxMalloc({sz}*sizeof(char));\n")
        fp.write(f"    if (mxGetString(prhs[{il}], in{il}_, {sz}) != 0) {{\n"
               f"        mw_err_txt_ = \"Invalid string argument\";\n"
               f"        goto mw_err_label;\n"
               f"    }}\n")
    fp.write("\n")


def _unpack_inputs_var(fp, ctx, args):
    for v in args:
        if v.iospec == 'o':
            continue
        if is_obj(v.tinfo):
            _cast_get_p(fp, ctx, v.basetype, v.input_label)
        elif is_array(v.tinfo):
            _unpack_input_array(fp, v)
        elif v.tinfo in (VT.scalar, VT.r_scalar, VT.p_scalar):
            il = v.input_label
            bt = v.basetype
            tp = _type_props(bt)
            fp.write(f"    if( mxGetClassID(prhs[{il}]) != {tp.scalar_class} )\n"
                   f"        mw_err_txt_ = \"Invalid scalar argument, {tp.scalar_class} expected\";\n"
                   f"    if (mw_err_txt_) goto mw_err_label;\n"
                   f"    in{il}_ = ({bt}) {tp.scalar_getter}(prhs[{il}], &mw_err_txt_);\n"
                   f"    if (mw_err_txt_)\n"
                   f"        goto mw_err_label;\n")
            if bt != "char":
                fp.write("\n")
        elif v.tinfo in (VT.cscalar, VT.zscalar,
                         VT.r_cscalar, VT.r_zscalar,
                         VT.p_cscalar, VT.p_zscalar):
            il = v.input_label
            bt = v.basetype
            tp = _type_props(bt)
            fp.write(f"    if( mxGetClassID(prhs[{il}]) != {tp.scalar_class} )\n"
                   f"        mw_err_txt_ = \"Invalid scalar argument, {tp.scalar_class} expected\";\n"
                   f"    if (mw_err_txt_) goto mw_err_label;\n")
            cs = _copier_suffix(bt)
            fp.write(f"    mxWrapGetScalar_{cs}{bt}(&in{il}_, prhs[{il}]);\n\n")
        elif v.tinfo == VT.string:
            _unpack_input_string(fp, v)
        elif v.tinfo == VT.mx:
            fp.write(f"    in{v.input_label}_ = prhs[{v.input_label}];\n\n")


def _unpack_inputs(fp, ctx, f):
    if f.thisv:
        _cast_get_p(fp, ctx, f.classv, 0)
    _unpack_inputs_var(fp, ctx, f.args)


# --- Step 5: Null-check objects/this ---

def _check_inputs(fp, args):
    for v in args:
        if v.iospec != 'o' and v.tinfo in (VT.obj, VT.r_obj):
            fp.write(f"    if (!in{v.input_label}_) {{\n"
                   f"        mw_err_txt_ = \"Argument {v.name} cannot be null\";\n"
                   f"        goto mw_err_label;\n"
                   f"    }}\n")


# --- Step 6: Allocate outputs ---

def _alloc_output(fp, ctx, args, return_flag):
    for v in args:
        if v.iospec == 'o':
            if v.devicespec != 'g':
                if not return_flag and is_obj(v.tinfo) and ctx.is_mxarray_type(v.basetype):
                    fp.write(f"    out{v.output_label}_ = mxWrapAlloc_{v.basetype}();\n")
                elif is_array(v.tinfo):
                    fp.write(f"    out{v.output_label}_ = ({v.basetype}*) mxMalloc({_alloc_size_expr(v.qual.args)}*sizeof({v.basetype}));\n")
                elif v.tinfo == VT.rarray:
                    fp.write(f"    out{v.output_label}_ = ({v.basetype}*) NULL;\n")
                elif v.tinfo == VT.string:
                    fp.write(f"    out{v.output_label}_ = (char*) mxMalloc({_alloc_size_expr(v.qual.args)}*sizeof(char));\n")
            if v.devicespec == 'g':
                da = v.qual.args
                ndims = 2 if len(da) == 2 else 1
                mtype = "mxCOMPLEX" if complex_tinfo(v) else "mxREAL"
                mxcid = basetype_to_mxclassid(v.basetype)
                cutype = basetype_to_cucomplex(v.basetype)
                if ndims == 2:
                    fp.write(f"    gpu_outdims{v.output_label}_[0] = dim{da[0].input_label}_; gpu_outdims{v.output_label}_[1] = dim{da[1].input_label}_;\n")
                else:
                    fp.write(f"    gpu_outdims{v.output_label}_[0] = dim{da[0].input_label}_;\n")
                fp.write(f"    mxGPUArray_out{v.output_label}_ = mxGPUCreateGPUArray({ndims}, gpu_outdims{v.output_label}_, {mxcid}, {mtype}, MX_GPU_DO_NOT_INITIALIZE);\n")
                fp.write(f"    out{v.output_label}_ = ({cutype} *)mxGPUGetData(mxGPUArray_out{v.output_label}_);\n\n")


def _alloc_outputs(fp, ctx, f):
    if not nullable_return(f):
        _alloc_output(fp, ctx, f.ret, True)
    _alloc_output(fp, ctx, f.args, False)


# --- Step 7: Profiler ---

def _record_call(fp, f):
    fp.write(f"    if (mexprofrecord_)\n"
           f"        mexprofrecord_[{f.id}]++;\n")


# --- Step 8: Make the call ---

def _make_call_args(fp, args, first):
    for v in args:
        if not first:
            fp.write(", ")
        n = vname(v)
        if v.tinfo in (VT.obj, VT.r_obj):
            fp.write(f"*{n}")
        elif v.tinfo == VT.mx and v.iospec == 'o':
            fp.write(f"plhs+{v.output_label}")
        elif v.tinfo in (VT.p_scalar, VT.p_cscalar, VT.p_zscalar):
            fp.write(f"&{n}")
        elif v.tinfo == VT.const:
            fp.write(v.name)
        else:
            fp.write(n)
        first = False


def _make_call_expr(fp, f):
    """Write the function call expression (without assignment/semicolon)."""
    if f.thisv:
        fp.write("in0_->")
    if f.funcv == "new":
        fp.write(f"new {f.classv}(")
    else:
        if f.fort:
            fp.write("MWF77_")
        fp.write(f"{f.funcv}(")
    _make_call_args(fp, f.args, True)
    fp.write(")")


def _make_stmt(fp, ctx, f):
    if f.thisv:
        fp.write("    if (!in0_) {\n"
               "        mw_err_txt_ = \"Cannot dispatch to NULL\";\n"
               "        goto mw_err_label;\n"
               "    }\n")

    if ctx.mw_generate_catch:
        fp.write("    try {\n    ")

    if f.ret:
        v = f.ret[0]
        if v.tinfo == VT.obj:
            if ctx.is_mxarray_type(v.basetype):
                fp.write(f"    plhs[0] = mxWrapSet_{v.basetype}(&(")
                _make_call_expr(fp, f)
                fp.write("));\n")
            else:
                fp.write(f"    out0_ = new {v.basetype}(")
                _make_call_expr(fp, f)
                fp.write(");\n")
        elif is_array(v.tinfo):
            fp.write(f"    plhs[0] = mxWrapReturn_{v.basetype}(")
            _make_call_expr(fp, f)
            fp.write(", ")
            args = v.qual.args
            if len(args) == 2:
                fp.write(f" dim{args[0].input_label}_, dim{args[1].input_label}_);\n")
            else:
                fp.write(f"{_alloc_size_expr(args)}, 1);\n")
        elif v.tinfo in (VT.scalar, VT.r_scalar, VT.cscalar, VT.r_cscalar, VT.zscalar, VT.r_zscalar):
            fp.write("    out0_ = ")
            _make_call_expr(fp, f)
            fp.write(";\n")
        elif v.tinfo == VT.string:
            fp.write("    plhs[0] = mxWrapStrncpy(")
            _make_call_expr(fp, f)
            fp.write(");\n")
        elif v.tinfo == VT.mx:
            fp.write("    plhs[0] = ")
            _make_call_expr(fp, f)
            fp.write(";\n")
        elif v.tinfo == VT.p_obj:
            if ctx.is_mxarray_type(v.basetype):
                fp.write(f"    plhs[0] = mxWrapSet_{v.basetype}(")
                _make_call_expr(fp, f)
                fp.write(");\n")
            else:
                fp.write("    out0_ = ")
                _make_call_expr(fp, f)
                fp.write(";\n")
        elif v.tinfo in (VT.p_scalar, VT.p_cscalar, VT.p_zscalar):
            fp.write(f"    plhs[0] = mxWrapReturn_{v.basetype}(")
            _make_call_expr(fp, f)
            fp.write(", 1, 1);\n")
        elif v.tinfo == VT.r_obj:
            if ctx.is_mxarray_type(v.basetype):
                fp.write(f"    plhs[0] = mxWrapSet_{v.basetype}(&(")
                _make_call_expr(fp, f)
                fp.write("));\n")
            else:
                fp.write("    out0_ = &(")
                _make_call_expr(fp, f)
                fp.write(");\n")
    else:
        fp.write("    ")
        _make_call_expr(fp, f)
        fp.write(";\n")

    if ctx.mw_generate_catch:
        fp.write(f"    }} catch(...) {{\n"
               f"        mw_err_txt_ = \"Caught C++ exception from {f.funcv}\";\n"
               f"    }}\n"
               f"    if (mw_err_txt_)\n"
               f"        goto mw_err_label;\n")


# --- Step 9: Marshal results ---

def _marshal_array(fp, v):
    il = v.input_label
    ol = v.output_label
    bt = v.basetype
    n = vname(v)

    if v.devicespec != 'g':
        da = v.qual.args
        mtype = "mxCOMPLEX" if complex_tinfo(v) else "mxREAL"
        ws = "    "
        is_single = _type_props(bt).is_single

        if v.tinfo == VT.rarray:
            ws = "        "
            fp.write(f"    if (out{ol}_ == NULL) {{\n")
            fp.write(f"        plhs[{ol}] = mxCreateDoubleMatrix(0,0, mxREAL);\n")
            fp.write(f"    }} else {{\n")

        if not da:
            # No dims — inout array
            if is_single:
                fp.write(f"{ws}plhs[{ol}] = mxCreateNumericMatrix(mxGetM(prhs[{il}]), mxGetN(prhs[{il}]), mxSINGLE_CLASS, {mtype});\n")
                fp.write(f"{ws}mxWrapCopy_single_{bt}(plhs[{ol}], in{il}_, ")
            else:
                fp.write(f"{ws}plhs[{ol}] = mxCreateDoubleMatrix(mxGetM(prhs[{il}]), mxGetN(prhs[{il}]), {mtype});\n")
                fp.write(f"{ws}mxWrapCopy_{bt}(plhs[{ol}], in{il}_, ")
            fp.write(f"mxGetM(prhs[{il}])*mxGetN(prhs[{il}])")
            fp.write(");\n")
        elif len(da) == 1:
            # 1D
            if is_single:
                fp.write(f"{ws}plhs[{ol}] = mxCreateNumericMatrix(dim{da[0].input_label}_, 1, mxSINGLE_CLASS, {mtype});\n")
                fp.write(f"{ws}mxWrapCopy_single_{bt}(plhs[{ol}], {n}, ")
            else:
                fp.write(f"{ws}plhs[{ol}] = mxCreateDoubleMatrix(dim{da[0].input_label}_, 1, {mtype});\n")
                fp.write(f"{ws}mxWrapCopy_{bt}(plhs[{ol}], {n}, ")
            fp.write(f"dim{da[0].input_label}_")
            fp.write(");\n")
        elif len(da) == 2:
            # 2D
            if is_single:
                fp.write(f"{ws}plhs[{ol}] = mxCreateNumericMatrix(dim{da[0].input_label}_, dim{da[1].input_label}_, mxSINGLE_CLASS, {mtype});\n")
                fp.write(f"{ws}mxWrapCopy_single_{bt}(plhs[{ol}], {n}, ")
            else:
                fp.write(f"{ws}plhs[{ol}] = mxCreateDoubleMatrix(dim{da[0].input_label}_, dim{da[1].input_label}_, {mtype});\n")
                fp.write(f"{ws}mxWrapCopy_{bt}(plhs[{ol}], {n}, ")
            fp.write(f"dim{da[0].input_label}_*dim{da[1].input_label}_")
            fp.write(");\n")
        else:
            # 3D+ — flatten to 1D
            sz = _alloc_size_expr(da)
            if is_single:
                fp.write(f"{ws}plhs[{ol}] = mxCreateNumericMatrix({sz}, 1, mxSINGLE_CLASS, {mtype});\n")
                fp.write(f"{ws}mxWrapCopy_single_{bt}(plhs[{ol}], {n}, ")
            else:
                fp.write(f"{ws}plhs[{ol}] = mxCreateDoubleMatrix({sz}, 1, {mtype});\n")
                fp.write(f"{ws}mxWrapCopy_{bt}(plhs[{ol}], {n}, ")
            fp.write(sz)
            fp.write(");\n")

        if v.tinfo == VT.rarray:
            fp.write("    }\n")

    # GPU marshal
    if v.devicespec == 'g':
        if v.iospec == 'b':
            fp.write(f"    plhs[{ol}] = prhs[{il}];\n")
        if v.iospec == 'o':
            fp.write(f"    plhs[{ol}] = mxGPUCreateMxArrayOnGPU(mxGPUArray_out{ol}_);\n")


def _marshal_result(fp, ctx, v, return_flag):
    n = vname(v)
    ol = v.output_label
    bt = v.basetype

    if is_obj(v.tinfo) and ctx.is_mxarray_type(bt):
        if not return_flag:
            fp.write(f"    plhs[{ol}] = mxWrapSet_{bt}({n});\n")
    elif is_obj(v.tinfo):
        fp.write(f"    plhs[{ol}] = mxWrapCreateP(out{ol}_, \"{bt}:%p\");\n")
    elif is_array(v.tinfo) or v.tinfo == VT.rarray:
        _marshal_array(fp, v)
    elif v.tinfo in (VT.scalar, VT.r_scalar, VT.p_scalar):
        _interleaved_branch(fp,
            f"    plhs[{ol}] = mxCreateDoubleMatrix(1, 1, mxREAL);\n"
            f"    *mxGetDoubles(plhs[{ol}]) = {n};\n",
            f"    plhs[{ol}] = mxCreateDoubleMatrix(1, 1, mxREAL);\n"
            f"    *mxGetPr(plhs[{ol}]) = {n};\n")
    elif v.tinfo in (VT.cscalar, VT.zscalar,
                     VT.r_cscalar, VT.r_zscalar,
                     VT.p_cscalar, VT.p_zscalar):
        _interleaved_branch(fp,
            f"    plhs[{ol}] = mxCreateDoubleMatrix(1, 1, mxCOMPLEX);\n"
            f"    mxGetComplexDoubles(plhs[{ol}])->real = real_{bt}({n});\n"
            f"    mxGetComplexDoubles(plhs[{ol}])->imag = imag_{bt}({n});\n",
            f"    plhs[{ol}] = mxCreateDoubleMatrix(1, 1, mxCOMPLEX);\n"
            f"    *mxGetPr(plhs[{ol}]) = real_{bt}({n});\n"
            f"    *mxGetPi(plhs[{ol}]) = imag_{bt}({n});\n")
    elif v.tinfo == VT.string:
        fp.write(f"    plhs[{ol}] = mxCreateString({n});\n")


def _marshal_results_var(fp, ctx, vars, return_flag):
    for v in vars:
        if v.iospec != 'i':
            _marshal_result(fp, ctx, v, return_flag)


def _marshal_results(fp, ctx, f):
    if not nullable_return(f):
        _marshal_results_var(fp, ctx, f.ret, True)
    _marshal_results_var(fp, ctx, f.args, False)


# --- Step 10: Dealloc ---

def _dealloc_var(fp, ctx, vars, return_flag):
    for v in vars:
        if v.devicespec != 'g':
            if is_array(v.tinfo) or v.tinfo == VT.string:
                if v.iospec == 'o':
                    fp.write(f"    if (out{v.output_label}_) mxFree(out{v.output_label}_);\n")
                elif v.iospec == 'b' or not (v.basetype == "double" or v.basetype == "float"):
                    fp.write(f"    if (in{v.input_label}_)  mxFree(in{v.input_label}_);\n")
            elif is_obj(v.tinfo) and ctx.is_mxarray_type(v.basetype):
                if v.iospec in ('i', 'b'):
                    fp.write(f"    if (in{v.input_label}_)  mxWrapFree_{v.basetype}(in{v.input_label}_);\n")
                elif v.iospec == 'o' and not return_flag:
                    fp.write(f"    if (out{v.output_label}_) mxWrapFree_{v.basetype}(out{v.output_label}_);\n")
        if v.devicespec == 'g':
            if v.iospec in ('i', 'b'):
                fp.write(f"    if (mxGPUArray_in{v.input_label}_)  mxGPUDestroyGPUArray(mxGPUArray_in{v.input_label}_);\n")
            if v.iospec == 'o':
                fp.write(f"    if (mxGPUArray_out{v.output_label}_)  mxGPUDestroyGPUArray(mxGPUArray_out{v.output_label}_);\n")


def _dealloc(fp, ctx, f):
    if not nullable_return(f):
        _dealloc_var(fp, ctx, f.ret, True)
    _dealloc_var(fp, ctx, f.args, False)


# ===================================================================
# Print a single MEX stub
# ===================================================================

def _print_c_comment(fp, f):
    fp.write(f"/* ---- {f.fname}: {f.line} ----\n")
    fp.write(f" * {print_func(f)}")
    # Preserve original behavior: only print first duplicate
    if f.same:
        fsame = f.same[0]
        fp.write(f" * Also at {fsame.fname}: {fsame.line}\n")
    fp.write(" */\n")


def _print_mex_stub(fp, ctx, f):
    _print_c_comment(fp, f)
    ids = id_string(ctx, f)
    fp.write(f"static const char* stubids{f.id}_ = \"{ids}\";\n\n")
    fp.write(f"void mexStub{f.id}(int nlhs, mxArray* plhs[],\n"
           f"              int nrhs, const mxArray* prhs[])\n"
           f"{{\n"
           f"    const char* mw_err_txt_ = 0;\n")
    _declare_args(fp, f)
    _unpack_dims(fp, f)
    _check_dims(fp, f.args)
    _unpack_inputs(fp, ctx, f)
    _check_inputs(fp, f.args)
    _alloc_outputs(fp, ctx, f)
    _record_call(fp, f)
    _make_stmt(fp, ctx, f)
    _marshal_results(fp, ctx, f)
    fp.write("\nmw_err_label:\n")
    _dealloc(fp, ctx, f)
    fp.write("    if (mw_err_txt_)\n"
           "        mexErrMsgTxt(mw_err_txt_);\n"
           "}\n\n")


# ===================================================================
# Print all stubs, dispatch table, mexFunction
# ===================================================================

def _print_mex_stubs(fp, ctx, funcs):
    for f in funcs:
        _print_mex_stub(fp, ctx, f)


def _make_profile_output(fp, funcs, printfunc):
    fp.write(f"        if (!mexprofrecord_)\n"
           f"            {printfunc}\"Profiler inactive\\n\");\n")
    for fc in funcs:
        fp.write(f"        {printfunc}\"%d calls to {fc.fname}:{fc.line}")
        # Preserve original behavior: only print first duplicate
        if fc.same:
            fp.write(f" ({fc.same[0].fname}:{fc.same[0].line})")
        fp.write(f"\\n\", mexprofrecord_[{fc.id}]);\n")


def _print_mex_else_cases(fp, funcs):
    for fc in funcs:
        fp.write(f"    else if (strcmp(id, stubids{fc.id}_) == 0)\n"
               f"        mexStub{fc.id}(nlhs,plhs, nrhs-1,prhs+1);\n")

    maxid = max_routine_id(funcs)
    fp.write(f"    else if (strcmp(id, \"*profile on*\") == 0) {{\n"
           f"        if (!mexprofrecord_) {{\n"
           f"            mexprofrecord_ = (int*) malloc({maxid+1} * sizeof(int));\n"
           f"            mexLock();\n"
           f"        }}\n"
           f"        memset(mexprofrecord_, 0, {maxid+1} * sizeof(int));\n"
           f"    }} else if (strcmp(id, \"*profile off*\") == 0) {{\n"
           f"        if (mexprofrecord_) {{\n"
           f"            free(mexprofrecord_);\n"
           f"            mexUnlock();\n"
           f"        }}\n"
           f"        mexprofrecord_ = NULL;\n"
           f"    }} else if (strcmp(id, \"*profile report*\") == 0) {{\n")
    _make_profile_output(fp, funcs, "mexPrintf(")
    fp.write(f"    }} else if (strcmp(id, \"*profile log*\") == 0) {{\n"
           f"        FILE* logfp;\n"
           f"        if (nrhs != 2 || mxGetString(prhs[1], id, sizeof(id)) != 0)\n"
           f"            mexErrMsgTxt(\"Must have two string arguments\");\n"
           f"        logfp = fopen(id, \"w+\");\n"
           f"        if (!logfp)\n"
           f"            mexErrMsgTxt(\"Cannot open log for output\");\n")
    _make_profile_output(fp, funcs, "fprintf(logfp, ")
    fp.write("        fclose(logfp);\n")
    fp.write("    } else\n"
           "        mexErrMsgTxt(\"Unknown identifier\");\n")


# ===================================================================
# Top-level: print_mex_init + print_mex_file
# ===================================================================

MWRAP_BANNER = (
    "/* --------------------------------------------------- */\n"
    "/* Automatically generated by mwrap                    */\n"
    "/* --------------------------------------------------- */\n\n"
)

MEX_BASE = (
    "/* ----\n"
    " */\n"
    "void mexFunction(int nlhs, mxArray* plhs[],\n"
    "                 int nrhs, const mxArray* prhs[])\n"
    "{\n"
    "    char id[1024];\n"
    "    if (nrhs == 0) {\n"
    "        mexPrintf(\"Mex function installed\\n\");\n"
    "        return;\n"
    "    }\n\n"
)

MEX_BASE_IF = (
    "    if (mxGetString(prhs[0], id, sizeof(id)) != 0)\n"
    "        mexErrMsgTxt(\"Identifier should be a string\");\n"
)


def print_mex_init(fp, ctx, support_text):
    """Write the MEX file header: banner + runtime support + complex/GPU includes."""
    fp.write(MWRAP_BANNER)
    fp.write(support_text)
    fp.write("\n")
    if ctx.mw_use_gpu:
        fp.write("#include <gpu/mxGPUArray.h>\n\n")
    if ctx.mw_use_c99_complex:
        mex_c99_complex(fp)
    elif ctx.mw_use_cpp_complex:
        mex_cpp_complex(fp)
        if ctx.mw_use_gpu:
            mex_gpucpp_complex(fp)


def print_mex_file(fp, ctx, funcs):
    """Write the rest of the MEX file: copiers, getters, stubs, dispatch."""
    if ctx.mw_use_int32_t or ctx.mw_use_int64_t or ctx.mw_use_uint32_t or ctx.mw_use_uint64_t:
        fp.write("#include <stdint.h>\n\n")
    mex_define_copiers(fp, ctx)
    mex_casting_getters(fp, ctx)

    if has_fortran(funcs):
        mex_define_fnames(fp, funcs)
        mex_fortran_decls(fp, funcs)

    _print_mex_stubs(fp, ctx, funcs)
    fp.write(MEX_BASE)
    fp.write("\n")
    if ctx.mw_use_gpu:
        fp.write("    mxInitGPU();\n")
    fp.write("\n")
    fp.write(MEX_BASE_IF)
    _print_mex_else_cases(fp, funcs)
    fp.write("}\n\n")
